namespace GossamerVR.Stripwise
{
    using System;
    using System.Collections;
    using System.Collections.Generic;
    using System.Collections.Specialized;
    using System.Diagnostics;
    using System.Linq;
    using System.Security.Principal;
    using System.Text;
    using UnityEngine;

    public class StripwiseLite : MVRScript
    {
        private JSONStorableFloat thresholdJSON;
        private JSONStorableFloat remapJSON;
        private JSONStorableFloat multiplierJSON;
        private JSONStorableFloat blendJSON;
        private JSONStorableFloat vertSizeJSON;
        private JSONStorableFloat xMinJSON;
        private JSONStorableFloat xMaxJSON;
        private JSONStorableFloat yMinJSON;
        private JSONStorableFloat yMaxJSON;
        private JSONStorableFloat zMinJSON;
        private JSONStorableFloat zMaxJSON;
        private JSONStorableFloat hMinJSON, hMaxJSON;
        private JSONStorableFloat cutoffPowerJSON;
        private JSONStorableFloat updateRateJSON;
        private JSONStorableFloat wMinJSON, wMaxJSON;
        //private JSONStorableBool deferUpdate;
        private JSONStorableAction resetCacheAction;
        private UIDynamicButton resetCacheButton;
        private JSONStorableBool yOnlyJSON;
        private List<ClothSimControl> sims = null;
        private Dictionary<ClothSimControl, float[]> jointCache = new Dictionary<ClothSimControl, float[]>();
        private Dictionary<ClothSimControl, Vector3[]> posCache = new Dictionary<ClothSimControl, Vector3[]>();
        private bool initialized;
        private bool isDirty = true;
        private float time = 0f;

        private Dictionary<JSONStorableFloat, float> lastValueLookup = new Dictionary<JSONStorableFloat, float>();
        private Vector3 minVals, maxVals;

        private Bounds bounds = new Bounds();
        private bool simsCleaned = false;
        private int updateDelay = 10;

        void OnEnable()
        {
            this.isDirty = true;
            this.updateDelay = 10;
            this.simsCleaned = false;
        }

        void OnDisable()
        {
            this.ClearAll();
        }

        public override void Init()
        {
            base.Init();

            this.thresholdJSON = new JSONStorableFloat("Max Sim Value Affected", 1f, this.UpdateSettings, 0f, 1f);
            this.RegisterFloat(this.thresholdJSON);

            this.remapJSON = new JSONStorableFloat("General Remap", 1f, this.UpdateSettings, 0f, 1f);
            this.RegisterFloat(this.remapJSON);

            this.multiplierJSON = new JSONStorableFloat("Constant Multiplier", 1f, this.UpdateSettings, 0f, 3f);
            this.RegisterFloat(this.multiplierJSON);

            this.xMinJSON = new JSONStorableFloat("X Cutoff Min", -1f, this.UpdateSettings, -1f, 1f, constrain: false);
            this.RegisterFloat(this.xMinJSON);

            this.xMaxJSON = new JSONStorableFloat("X Cutoff Max", 1f, this.UpdateSettings, -1f, 1f, constrain: false);
            this.RegisterFloat(this.xMaxJSON);

            this.yMinJSON = new JSONStorableFloat("Y Cutoff Min", 0f, this.UpdateSettings, 0f, 2.5f, constrain: false);
            this.RegisterFloat(this.yMinJSON);

            this.yMaxJSON = new JSONStorableFloat("Y Cutoff Max", 2.5f, this.UpdateSettings, 0f, 2.5f, constrain: false);
            this.RegisterFloat(this.yMaxJSON);

            this.zMinJSON = new JSONStorableFloat("Z Cutoff Min", -.6f, this.UpdateSettings, -.6f, .6f, constrain: false);
            this.RegisterFloat(this.zMinJSON);

            this.zMaxJSON = new JSONStorableFloat("Z Cutoff Max", .6f, this.UpdateSettings, -.6f, .6f, constrain: false);
            this.RegisterFloat(this.zMaxJSON);

            this.hMinJSON = new JSONStorableFloat("Bottom Effect Bounds", 0f, this.UpdateSettings, 0f, 2.5f, constrain: false);
            this.RegisterFloat(this.hMinJSON);

            this.hMaxJSON = new JSONStorableFloat("Top Effect Bounds", 2.5f, this.UpdateSettings, 0f, 2.5f, constrain: false);
            this.RegisterFloat(this.hMaxJSON);

            this.wMinJSON = new JSONStorableFloat("Left Effect Bounds", -1.25f, this.UpdateSettings, -1.25f, 1.25f, constrain: false);
            this.RegisterFloat(this.wMinJSON);

            this.wMaxJSON = new JSONStorableFloat("Right Effect Bounds", 1.25f, this.UpdateSettings, -1.25f, 1.25f, constrain: false);
            this.RegisterFloat(this.wMaxJSON);


            this.cutoffPowerJSON = new JSONStorableFloat("Falloff", 0.5f, this.UpdateSettings, 0.01f, 4f, constrain: true);
            this.RegisterFloat(this.cutoffPowerJSON);

            this.updateRateJSON = new JSONStorableFloat("Update Rate", 10f, 0f, 60f, constrain: false);

            this.yOnlyJSON = new JSONStorableBool("Ignore X and Z (faster)", false, (bool x) => { this.isDirty = true; });

            //this.deferUpdate = new JSONStorableBool("Deferred Update", false, (bool x) => {this.isDirty = true; }); // not ready yet

            this.resetCacheAction = new JSONStorableAction("Reset Sim Cache", this.ResetCache);            
        }


        public override void InitUI()
        {
            base.InitUI();
            if (this.UITransform != null)
            {
                this.CreateSlider(this.thresholdJSON, false);
                this.CreateSlider(this.remapJSON, true);
                this.CreateSlider(this.multiplierJSON, false);
                this.CreateSlider(this.cutoffPowerJSON, true);

                this.CreateSlider(this.xMinJSON, false);
                this.CreateSlider(this.xMaxJSON, true);
                this.CreateSlider(this.yMinJSON, false);
                this.CreateSlider(this.yMaxJSON, true);
                this.CreateSlider(this.zMinJSON, false);
                this.CreateSlider(this.zMaxJSON, true);

                this.CreateSlider(this.hMinJSON, false);
                this.CreateSlider(this.hMaxJSON, true);

                this.CreateSlider(this.wMinJSON, false);
                this.CreateSlider(this.wMaxJSON, true);

                this.CreateSlider(this.updateRateJSON, false);
                this.CreateToggle(this.yOnlyJSON, true);

                //this.CreateToggle(this.deferUpdate, true);
                this.resetCacheButton = this.CreateButton("Reset Sim Cache", false);
                this.resetCacheAction.RegisterButton(this.resetCacheButton);
                this.initialized = true;
                this.isDirty = true;
            }
        }

        private void ClearAll()
        {
            if (this.sims != null)
            {
                for (int s = 0; s < this.sims.Count; s++)
                {
                    var sim = this.sims[s];
                    if (sim != null)
                    {
                        sim.Reset();
                    }
                }
            }
            this.sims = null;
            this.updateDelay = 10;
            this.jointCache.Clear();
            this.posCache.Clear();
            this.isDirty = false;
            this.simsCleaned = false;
        }

        /// <summary>
        /// Finds and loads the sims within the character.
        /// if not initialized or we haven't had the character atom defined yet,
        /// we ensure we'll try again by nullifying any list we currently have.
        /// </summary>
        private void LoadSims()
        {
            if (this.initialized && this.containingAtom != null)
            {
                this.sims = new List<ClothSimControl>(this.containingAtom.GetComponentsInChildren<ClothSimControl>());
            }
            else
            {
                this.sims = null;
            }
            this.isDirty = true;
            this.simsCleaned = false;
        }

        /// <summary>
        /// this brutish method wipes all memory and resets the clothes
        /// then applies the settings again. It's like turning it off and on.
        /// It's the only way currently the system will learn about new clothes.
        /// </summary>
        private void ResetCache()
        {
            this.ClearAll();
            this.LoadSims();

            // to do: add a listener for clothing items appearing and disappearing, instead of this.
        }


        /// <summary>
        /// Callback for when a slider or UI element changes.
        /// This will set the dirty flag if the change was significant enough to warrant it.
        /// this significance threshold helps allow for timeline to animate sliders
        /// without triggering a refresh if they don't change.
        /// </summary>
        private void UpdateSettings(JSONStorableFloat changedFloat)
        {
            if (!this.lastValueLookup.ContainsKey(changedFloat))
            {
                this.lastValueLookup[changedFloat] = changedFloat.val;
                this.isDirty = true;
            }
            else if (Mathf.Abs(this.lastValueLookup[changedFloat] - changedFloat.val) > 0.01f)
            {
                this.isDirty = true;
            }

            if (this.isDirty)
            {
                this.bounds.min = new Vector3(this.wMinJSON.val, this.hMinJSON.val, -2f);
                this.bounds.max = new Vector3(this.wMaxJSON.val, this.hMaxJSON.val, 2f);
            }

        }


        private void Update()
        {
            if (this.initialized)
            {
                this.time += Time.deltaTime;
                if (this.updateDelay <= 0 && this.isDirty && (this.updateRateJSON.val < 1f || this.time * this.updateRateJSON.val > 1f))
                {
                    this.isDirty = false;
                    this.CalculateAndCommitProperties();
                    this.time = 0f;
                }
                else this.updateDelay--;
            }
        }

        private void CalculateAndCommitProperties()
        {

            this.minVals.x = this.xMinJSON.val;
            this.minVals.y = this.yMinJSON.val;
            this.minVals.z = this.zMinJSON.val;

            this.maxVals.x = this.xMaxJSON.val;
            this.maxVals.y = this.yMaxJSON.val;
            this.maxVals.z = this.zMaxJSON.val;

            float cutoff = this.cutoffPowerJSON.val;
            float remapValue = this.remapJSON.val;
            float multiplier = this.multiplierJSON.val;
            bool ignoreXZ = this.yOnlyJSON.val;
            float[] rigidCache;
            Vector3[] pset;
            if (this.sims == null)
            {
                this.LoadSims();
                if (this.sims == null) return; // inelegant early day bandaid
            }

            for (int s = 0; s < this.sims.Count; s++)
            {
                var sim = this.sims[s];
                if (!this.simsCleaned)
                {
                    // if we've freshly loaded clothing ,we want to weed out any invalid items.
                    if (sim == null || sim.name.ToLower().Contains("necklace") || sim.clothSettings == null || sim.clothSettings.Runtime == null || sim.clothSettings.Runtime.Particles == null || sim.clothSettings.Runtime.PointJoints == null)
                    {
                        // we shouldn't even consider non-sim clothing.
                        if (sim != null) this.Log("Stripwise detected missing or invalid cloth settings, ignoring & removing " + sim.name);
                        this.sims.RemoveAt(s);
                        s--;
                        continue;
                    }
                }

                if (!sim.gameObject.activeInHierarchy || !sim.clothSettings.enabled)
                {
                    // we don't want to waste resources on clothing not active or with sim disabled.
                    continue;
                }

                if (!this.jointCache.ContainsKey(sim))
                {
                    sim.Reset();
                    sim.clothSettings.BreakEnabled = false;
                }

                /// pull the simulation points so the GPU marshals the data for us to use.
                sim.clothSettings.Runtime.Particles.PullData();
                sim.clothSettings.Runtime.PointJoints.PullData();
                if (!this.jointCache.TryGetValue(sim, out rigidCache))
                {
                    this.Log("Stripwise adding new sim clothing: " + sim.name);
                    rigidCache = new float[sim.clothSettings.Runtime.PointJoints.Count];
                    pset = new Vector3[sim.clothSettings.Runtime.Particles.Count];
                    this.jointCache[sim] = rigidCache;
                    this.posCache[sim] = pset;
                    for (int i = 0; i < sim.clothSettings.Runtime.PointJoints.Count; i++)
                    {
                        rigidCache[i] = sim.clothSettings.Runtime.PointJoints.Data[i].Rigidity;
                        pset[i] = sim.clothSettings.Runtime.Particles.Data[i].Position;
                    }
                }
                else
                {
                    pset = this.posCache[sim];
                }

                int index = 0;
                float xRigid = 1f;
                float yRigid = 1f;
                float zRigid = 1f;
                float rigidity = 1f;

                Vector3 pos;

                bool log = false;
                for (index = 0; index < sim.clothSettings.Runtime.PointJoints.Count; index++)
                {
                    pos = pset[index];
                    if (rigidCache[index] <= this.thresholdJSON.val && this.bounds.Contains(pos))
                    {

                        rigidity = Mathf.InverseLerp(0f, remapValue, rigidCache[index]);
                        xRigid = ignoreXZ ? 1f : Mathf.Pow(this.GetCutoffValue(pos.x, this.minVals.x, this.maxVals.x), cutoff);
                        yRigid = Mathf.Pow(this.GetCutoffValue(pos.y, this.minVals.y, this.maxVals.y), cutoff);
                        zRigid = ignoreXZ ? 1f : Mathf.Pow(this.GetCutoffValue(pos.z, this.minVals.z, this.maxVals.z), cutoff);
                        rigidity *= multiplier * xRigid * yRigid * zRigid;

                        sim.clothSettings.Runtime.PointJoints.Data[index].Rigidity = rigidity;
                    }
                    else
                    {
                        sim.clothSettings.Runtime.PointJoints.Data[index].Rigidity = rigidCache[index];
                    }
                }

                /// push the updated simulation points back so the data is available for the GPU
                sim.clothSettings.Runtime.PointJoints.PushData();
                sim.clothSettings.Runtime.Particles.PushData();

            }
            this.simsCleaned = true;
        }

        /// <summary>
        /// Returns a value between 0 and 1 based on whether or not the input is between min and max.
        /// allows min and max to be reversed.
        /// </summary>                
        private float GetCutoffValue(float input, float min, float max)
        {
            return Mathf.Clamp01((input - min) * (max - input));

        }



        private void Log(string msg)
        {
            SuperController.LogMessage(msg);
        }

    }
}